
/*  RAxML-VI-HPC (version 2.0) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright June 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */

#include <math.h>
#include <unistd.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "axml.h"


#define SHFT(a,b,c,d)                (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b)                    ((b) > 0.0 ? fabs(a) : -fabs(a))

double MNBRAK_GOLD =    1.618034;
double MNBRAK_TINY =      1.e-20;
double MNBRAK_GLIMIT =     100.0;
double BRENT_ITMAX =         100;
double BRENT_ZEPS  =      1.e-5;
double BRENT_CGOLD =   0.3819660;


extern int optimizeRatesInvocations;
extern int optimizeRateCategoryInvocations;
extern int optimizeAlphaInvocations;


/*********************FUNCTIONS FOOR EXACT MODEL OPTIMIZATION UNDER GTRGAMMA ***************************************/

double evaluateRate(tree *tr, int i, double rate, analdef *adef, int model)
{
  tr->initialRates[model * 5 + i] = rate;

  initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);

  if(adef->useMultipleModel)
    {
      return -partitionLikelihood(tr, model);
    }
  else
    {
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start); 
      
      return (-tr->likelihood);
    }
}

double evaluateAlpha(tree *tr, double alpha, int model, analdef *adef)
{
   tr->alphas[model] = alpha;    
   makeGammaCats(tr, model); 

   if(adef->useMultipleModel)
     {
       return -partitionLikelihood(tr, model);
     }
   else
     {
       initrav(tr, tr->start);
       initrav(tr, tr->start->back);
       evaluate(tr, tr->start);
      
       return (-tr->likelihood);
     }
}


int brakAlpha(double *param, double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double lim_inf, double lim_sup, tree *tr, int model, analdef *adef)
{
   double ulim,u,r,q,fu,dum;

   u = 0.0;

   *param = *ax;
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fa = evaluateAlpha(tr, *param, model, adef);

   *param = *bx;
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fb = evaluateAlpha(tr, *param, model, adef);

   if (*fb > *fa) 
     {
       SHFT(dum,*ax,*bx,dum)
       SHFT(dum,*fb,*fa,dum)
     }

   *cx=(*bx)+MNBRAK_GOLD*(*bx-*ax);
   *param = fabs(*cx);
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fc = evaluateAlpha(tr, *param, model, adef); 

   while (*fb > *fc) 
     {        
       if(*ax > lim_sup) *ax = lim_sup;
       if(*ax < lim_inf) *ax = lim_inf;
       if(*bx > lim_sup) *bx = lim_sup;
       if(*bx < lim_inf) *bx = lim_inf;
       if(*cx > lim_sup) *cx = lim_sup;
       if(*cx < lim_inf) *cx = lim_inf;
       if(u   > lim_sup) u   = lim_sup;
       if(u   < lim_inf) u   = lim_inf;

       r=(*bx-*ax)*(*fb-*fc);
       q=(*bx-*cx)*(*fb-*fa);
       u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
               (2.0*SIGN(MAX(fabs(q-r),MNBRAK_TINY),q-r));
       ulim=(*bx)+MNBRAK_GLIMIT*(*cx-*bx);
       
       if ((*bx-u)*(u-*cx) > lim_inf) 
	 {
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateAlpha(tr, *param, model, adef);
	   if (fu < *fc) 
	     {
	       *ax=(*bx);
	       *bx=u;
	       *fa=(*fb);
	       *fb=fu;
	       (*ax)=fabs(*ax);
	       (*bx)=fabs(*bx);
	       (*cx)=fabs(*cx);
	       return(0);
	     } 
	   else if (fu > *fb) 
	     {
	       *cx=u;
	       *fc=fu;	
	       (*ax)=fabs(*ax);
	       (*bx)=fabs(*bx);
	       (*cx)=fabs(*cx);
	       return(0);
	     }
	   u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu= evaluateAlpha(tr, *param, model, adef);
	 } 
       else if ((*cx-u)*(u-ulim) > lim_inf) 
	 {
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateAlpha(tr, *param, model, adef);
	   if (fu < *fc) 
	     {
	       SHFT(*bx,*cx,u,*cx+MNBRAK_GOLD*(*cx-*bx))
	       *param = fabs(u); 
	       SHFT(*fb,*fc,fu, evaluateAlpha(tr, *param, model, adef))
	     }
	 } 
       else if ((u-ulim)*(ulim-*cx) >= lim_inf) 
	 {
	   u=ulim;
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateAlpha(tr, *param, model, adef);
	 } 
       else 
	 {
	   u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateAlpha(tr, *param, model, adef);
	 }
       SHFT(*ax,*bx,*cx,u)
       SHFT(*fa,*fb,*fc,fu)


     }
   (*ax)=fabs(*ax);
   (*bx)=fabs(*bx);
   (*cx)=fabs(*cx);
   return(0);
}

double brentAlpha(double fb, double *param, double ax, double bx, double cx, double tol, double *xmin, tree *tr, int n_iter_max, int model, analdef *adef)
{
  int iter;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e=0.0;
  double init_loglk, max_loglk;
  double bestx;

  d=0.0;
  a=((ax < cx) ? ax : cx);
  b=((ax > cx) ? ax : cx);
  x=w=v=bx;
  *param=bx;
 
  fw=fv=fx= fb;

  if(adef->useMultipleModel)
    init_loglk = partitionLikelihood(tr, model);
  else
    init_loglk = tr->likelihood;

  max_loglk = unlikely;
  bestx = bx;

  for(iter=1;iter<=BRENT_ITMAX;iter++) 
    {
      xm=0.5*(a+b);
      tol2=2.0*(tol1=tol*fabs(x)+BRENT_ZEPS);
      if(fabs(x-xm) <= (tol2-0.5*(b-a))) 
	{	  
	  *xmin=x;
	  return -fx;
	}
      
      if(fabs(e) > tol1) 
	{
	  r=(x-w)*(fx-fv);
	  q=(x-v)*(fx-fw);
	  p=(x-v)*q-(x-w)*r;
	  q=2.0*(q-r);
	  if(q > 0.0) p = -p;
	  q=fabs(q);
	  etemp=e;
	  e=d;
	  if(fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
              {
                  d=BRENT_CGOLD*(e=(x >= xm ? a-x : b-x));
              }
	  else
              {
                  d=p/q;
                  u=x+d;
                  if (u-a < tol2 || b-u < tol2)
                      d=SIGN(tol1,xm-x);
              }
        }
      else
          {
              d=BRENT_CGOLD*(e=(x >= xm ? a-x : b-x));
          }

      u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
      *param=u;      
      fu = evaluateAlpha(tr, *param, model, adef);
      
      if(fu > max_loglk)
          {
              max_loglk = fu;
              bestx = u;
          }

      if(fu <= fx) 
	{
	  if(iter > n_iter_max) 
	    {	      		  
	      return fu;
	    }
	  if(u >= x) a=x; else b=x;
	  SHFT(v,w,x,u)
	    SHFT(fv,fw,fx,fu)
	    } 
      else
	{
	  if (u < x) a=u; else b=u;
	  if (fu <= fw || w == x) 
	    {
	      v=w;
	      w=u;
	      fv=fw;
            fw=fu;
	    } 
	  else if (fu <= fv || v == x || v == w) {
            v=u;
            fv=fu;
	  }
	}
    }
  printf("\n. Too many iterations in BRENT !");
  exit(-1);
  return(-1);
  /* Not Reached ??  *xmin=x;   */
  /* Not Reached ??  return fx; */
}


int brakRates(double *param, double *ax, double *bx, double *cx, double *fa, double *fb, double *fc, double lim_inf, double lim_sup, tree *tr, int i, analdef *adef, int model)
{
   double ulim,u,r,q,fu,dum;

   u = 0.0;

   *param = *ax;
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fa = evaluateRate(tr, i, *param, adef, model);

   *param = *bx;
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fb = evaluateRate(tr, i, *param, adef, model);

   if (*fb > *fa) 
     {
       SHFT(dum,*ax,*bx,dum)
       SHFT(dum,*fb,*fa,dum)
     }

   *cx=(*bx)+MNBRAK_GOLD*(*bx-*ax);
   *param = fabs(*cx);
   if(*param > lim_sup) *param = lim_sup;
   if(*param < lim_inf) *param = lim_inf;
   *fc =  evaluateRate(tr, i, *param, adef, model);

   while (*fb > *fc) 
     {        
       if(*ax > lim_sup) *ax = lim_sup;
       if(*ax < lim_inf) *ax = lim_inf;
       if(*bx > lim_sup) *bx = lim_sup;
       if(*bx < lim_inf) *bx = lim_inf;
       if(*cx > lim_sup) *cx = lim_sup;
       if(*cx < lim_inf) *cx = lim_inf;
       if(u   > lim_sup) u   = lim_sup;
       if(u   < lim_inf) u   = lim_inf;

       r=(*bx-*ax)*(*fb-*fc);
       q=(*bx-*cx)*(*fb-*fa);
       u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
               (2.0*SIGN(MAX(fabs(q-r),MNBRAK_TINY),q-r));
       ulim=(*bx)+MNBRAK_GLIMIT*(*cx-*bx);
       
       if ((*bx-u)*(u-*cx) > lim_inf) 
	 {
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateRate(tr, i, *param, adef, model);
	   if (fu < *fc) 
	     {
	       *ax=(*bx);
	       *bx=u;
	       *fa=(*fb);
	       *fb=fu;
	       (*ax)=fabs(*ax);
	       (*bx)=fabs(*bx);
	       (*cx)=fabs(*cx);
	       return(0);
	     } 
	   else if (fu > *fb) 
	     {
	       *cx=u;
	       *fc=fu;	
	       (*ax)=fabs(*ax);
	       (*bx)=fabs(*bx);
	       (*cx)=fabs(*cx);
	       return(0);
	     }
	   u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu= evaluateRate(tr, i, *param, adef, model);
	 } 
       else if ((*cx-u)*(u-ulim) > lim_inf) 
	 {
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateRate(tr, i, *param, adef, model);
	   if (fu < *fc) 
	     {
	       SHFT(*bx,*cx,u,*cx+MNBRAK_GOLD*(*cx-*bx))
	       *param = fabs(u); 
	       SHFT(*fb,*fc,fu, evaluateRate(tr, i, *param, adef, model))
	     }
	 } 
       else if ((u-ulim)*(ulim-*cx) >= lim_inf) 
	 {
	   u=ulim;
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateRate(tr, i, *param, adef, model);
	 } 
       else 
	 {
	   u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
	   *param = fabs(u);
	   if(*param > lim_sup) {*param = u = lim_sup;}
	   if(*param < lim_inf) {*param = u = lim_inf;}
	   fu = evaluateRate(tr, i, *param, adef, model);
	 }
       SHFT(*ax,*bx,*cx,u)
       SHFT(*fa,*fb,*fc,fu)


     }
   (*ax)=fabs(*ax);
   (*bx)=fabs(*bx);
   (*cx)=fabs(*cx);
   return(0);
}

double brentRates(double fb, double *param, double ax, double bx, double cx, double tol, double *xmin, tree *tr, int n_iter_max, int i, analdef *adef, int model)
{
  int iter;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e=0.0;
  double init_loglk, max_loglk;
  double bestx;

  d=0.0;
  a=((ax < cx) ? ax : cx);
  b=((ax > cx) ? ax : cx);
  x=w=v=bx;
  *param=bx;
 
  fw=fv=fx= fb;

  if(adef->useMultipleModel)
    init_loglk = partitionLikelihood(tr, model);
  else
    init_loglk = tr->likelihood;

  max_loglk = unlikely;
  bestx = bx;

  for(iter=1;iter<=BRENT_ITMAX;iter++) 
    {
      xm=0.5*(a+b);
      tol2=2.0*(tol1=tol*fabs(x)+BRENT_ZEPS);
      if(fabs(x-xm) <= (tol2-0.5*(b-a))) 
	{	  
	  *xmin=x;
	  return -fx;
	}
      
      if(fabs(e) > tol1) 
	{
	  r=(x-w)*(fx-fv);
	  q=(x-v)*(fx-fw);
	  p=(x-v)*q-(x-w)*r;
	  q=2.0*(q-r);
	  if(q > 0.0) p = -p;
	  q=fabs(q);
	  etemp=e;
	  e=d;
	  if(fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
              {
                  d=BRENT_CGOLD*(e=(x >= xm ? a-x : b-x));
              }
	  else
              {
                  d=p/q;
                  u=x+d;
                  if (u-a < tol2 || b-u < tol2)
                      d=SIGN(tol1,xm-x);
              }
        }
      else
          {
              d=BRENT_CGOLD*(e=(x >= xm ? a-x : b-x));
          }

      u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
      *param=u;      
      fu = evaluateRate(tr, i, *param, adef, model);
      
      if(fu > max_loglk)
          {
              max_loglk = fu;
              bestx = u;
          }

      if(fu <= fx) 
	{
	  if(iter > n_iter_max) 
	    {	      		  
	      return fu;
	    }
	  if(u >= x) a=x; else b=x;
	  SHFT(v,w,x,u)
	    SHFT(fv,fw,fx,fu)
	    } 
      else
	{
	  if (u < x) a=u; else b=u;
	  if (fu <= fw || w == x) 
	    {
	      v=w;
	      w=u;
	      fv=fw;
            fw=fu;
	    } 
	  else if (fu <= fv || v == x || v == w) {
            v=u;
            fv=fu;
	  }
	}
    }
  printf("\n. Too many iterations in BRENT !");
  exit(-1);
  return(-1);
  /* Not Reached ??  *xmin=x;   */
  /* Not Reached ??  return fx; */
}




void optAlpha(tree *tr, analdef *adef, double modelEpsilon, int model)
{
  double param, a, b, c, fa, fb, fc, x;
  double lim_inf = ALPHA_MIN;
  double lim_sup = ALPHA_MAX;
  double startLH;
  double startAlpha = tr->alphas[model];
  double endAlpha;

  if(adef->useMultipleModel)
    startLH = partitionLikelihood(tr, model);
  else
    startLH = tr->likelihood;

  a = tr->alphas[model] + 0.1;
  b = tr->alphas[model] - 0.1; 
  
  if(b < lim_inf) b = lim_inf;
  
  brakAlpha(&param, &a, &b, &c, &fa, &fb, &fc, lim_inf, lim_sup, tr, model, adef);
  /*  printf("brackets %f %f %f, %f %f %f\n", a, b, c, fa, fb, fc);*/
  /*printf("brak alpha %d\n", ea);*/
      
  endAlpha = brentAlpha(fb, &param, a, b, c, modelEpsilon, &x,  tr, 100, model, adef);   
  /*  printf("brent alpha %d\n", endAlpha);  */

  if(startLH > endAlpha)
    {
      tr->alphas[model] = startAlpha;
      evaluateAlpha(tr, tr->alphas[model], model, adef);
    }

#ifdef DEBUG_MODEL
  if(adef->useMultipleModel)
    printf("Alpha Partition %d: %f -> %f\n", model, startLH, endAlpha);
  else
    printf("Alpha: %f -> %f\n", startLH, endAlpha);
#endif
}

double optRates(tree *tr, analdef *adef, double modelEpsilon, int model)
{
  int i;
  double param, a, b, c, fa, fb, fc, x;
  double lim_inf = RATE_MIN;
  double lim_sup = RATE_MAX;
  double startRates[5];
  double startLH, start2;
  double endLH;

  if(adef->useMultipleModel)
    startLH = partitionLikelihood(tr, model);
  else
    startLH = tr->likelihood;

  start2 = startLH;
    
  for(i = 0; i < 5; i++)
    {
      startRates[i] = tr->initialRates[model * 5 + i];

      a = tr->initialRates[model * 5 + i] + 0.1;
      b = tr->initialRates[model * 5 + i] - 0.1;
      
      if(b < lim_inf) b = lim_inf;
      
      brakRates(&param, &a, &b, &c, &fa, &fb, &fc, lim_inf, lim_sup, tr, i, adef, model);
      
      /*printf("brackets %f %f %f, %f %f %f\n", a, b, c, fa, fb, fc);*/
      /*printf("brak rates %d\n", er);*/
      
      endLH = brentRates(fb, &param, a, b, c, modelEpsilon, &x,  tr, 100, i, adef, model); 
      /*      printf("brent rates %d\n", endLH);*/
   

      if(startLH > endLH)
	{
	  tr->initialRates[model * 5 + i] = startRates[i];
	  initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model); 

	  if(adef->useMultipleModel)	    
	    partitionLikelihood(tr, model);	    
	  else
	    {
	      initrav(tr, tr->start);
	      initrav(tr, tr->start->back);
	      evaluate(tr, tr->start); 
	    }
	}
      else
	startLH = endLH;                     
    }
#ifdef DEBUG_MODEL
  if(adef->useMultipleModel)
    printf("Rates Partition %d: %f -> %f\n", model, start2, endLH);
  else
    printf("Rates: %f -> %f\n", start2, endLH);
#endif
  return endLH;
}


void resetBranches(tree *tr)
{
  nodeptr  p, q;
  int  nodes;
  
  nodes = tr->mxtips  +  3 * (tr->mxtips - 2);
  p = tr->nodep[1];
  while (nodes-- > 0) 
    {
      p->z = defaultz; 
      q = p->next;
      while(q != p)
	{
	  q->z = defaultz;
	  q = q->next;
	}
      p++;
    }
}



void modOpt(tree *tr, analdef *adef)
{
  double currentLikelihood, intermediateLikelihood;
  int i; 
  double modelEpsilon = MODEL_EPSILON;
  double optTime = gettime();
  int model;
  double *partialLikelihoods = (double *)malloc(tr->NumberOfModels * sizeof(double));

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      tr->initialRates[model * 5] = 0.5;
      tr->initialRates[model * 5 + 1] = 0.5;
      tr->initialRates[model * 5 + 2] = 0.5;
      tr->initialRates[model * 5 + 3] = 0.5;
      tr->initialRates[model * 5 + 4] = 0.5;      
      tr->alphas[model] = 1.0;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);
      makeGammaCats(tr, model); 
    }

  resetBranches(tr);
  tr->start = tr->nodep[1];

  /* no need for individual models here, just an init on params equal for all partitions*/
  initrav(tr, tr->start);
  initrav(tr, tr->start->back);
  evaluate(tr, tr->start); 

  treeEvaluate(tr, 1.0);
  /*printf("Starting with %f node %d %d\n", tr->likelihood, tr->start->number, tr->start->back->number);*/

  


  do
    {
      currentLikelihood = tr->likelihood;
                  
      if(adef->model == M_GTRGAMMA)
	{	       	    
	  for(model = 0; model < tr->NumberOfModels; model++)
	    partialLikelihoods[model] = optRates(tr, adef, modelEpsilon, model);   	
	  for(model = 0; model < tr->NumberOfModels; model++)
	    optAlpha(tr, adef, modelEpsilon, model);       	 	
	}
      else
	{
	  printf("Internal error this function should only be called with GTR+GAMMA\n");
	  exit(-1);
	} 

      if(tr->NumberOfModels > 1)
	{
	   tr->fracchange = 0;
	   for(model = 0; model < tr->NumberOfModels; model++)
	     tr->fracchange += tr->fracchanges[model] * abs(partialLikelihoods[model]);
	   tr->fracchange /= abs(tr->likelihood);
	}

      /*printf("FRACCHANGE %f\n", tr->fracchange);*/
      
      treeEvaluate(tr, 0.25);
           
      modelEpsilon /= 10.0;
      if(modelEpsilon < LIKELIHOOD_EPSILON)
	modelEpsilon = LIKELIHOOD_EPSILON;

      /*      printf("%f %f\n", gettime() - optTime, tr->likelihood);*/
    }
  while(fabs(currentLikelihood - tr->likelihood) > adef->likelihoodEpsilon);

  free(partialLikelihoods);
}

/*********************FUNCTIONS FOOR EXACT MODEL OPTIMIZATION UNDER GTRGAMMA ***************************************/


/*********************FUNCTIONS FOR APPROXIMATE MODEL OPTIMIZATION ***************************************/

void optimizeAlphaMULT(tree *tr, int model, analdef *adef)
{
  int k; 
  double currentTT, startLikelihood, maxLikelihoodMinus,
    maxLikelihoodPlus, maxTTPlus, maxTTMinus, spacing,
    tree_likelihood;
  boolean finish = FALSE;

  spacing = 0.5/((double)optimizeAlphaInvocations);

  currentTT = tr->alphas[model];
  
  tree_likelihood = partitionLikelihood(tr, model);
 
  maxTTPlus = currentTT;
  maxTTMinus = currentTT;

  startLikelihood = tree_likelihood;
  maxLikelihoodMinus = tree_likelihood;
  maxLikelihoodPlus = tree_likelihood;
      
  k = 1;       

  while(!finish /*tree_likelihood >= maxLikelihoodMinus*/ && ((currentTT - spacing * k) > ALPHA_MIN)) 
    {           
      tr->alphas[model] = currentTT - spacing * k;     
                
      makeGammaCats(tr, model);
     
      tree_likelihood = partitionLikelihood(tr, model);
           
      if(tree_likelihood > maxLikelihoodMinus)
	{
	  finish = (fabs(tree_likelihood - maxLikelihoodMinus) < adef->likelihoodEpsilon);

	  maxLikelihoodMinus = tree_likelihood;
	  maxTTMinus = currentTT - spacing * k;
	}    
      else
	finish = TRUE;

      k++;
    }

  finish = FALSE;
  k = 1;
  tree_likelihood = startLikelihood;

  while(!finish /*tree_likelihood >= maxLikelihoodPlus*/ && ((currentTT + spacing * k) < ALPHA_MAX)) 
    {      
      tr->alphas[model] = currentTT + spacing * k;     
      makeGammaCats(tr, model);
      
      tree_likelihood = partitionLikelihood(tr, model);     

      if(tree_likelihood > maxLikelihoodPlus)
	{
	  finish = (fabs(tree_likelihood - maxLikelihoodPlus) < adef->likelihoodEpsilon);

	  maxLikelihoodPlus = tree_likelihood;
	  maxTTPlus = currentTT + spacing * k;
	}	 
      else
	finish = TRUE;

      k++;
    }  

  if(maxLikelihoodPlus > startLikelihood || maxLikelihoodMinus > startLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)	
	tr->alphas[model] = maxTTPlus;	
      else	
	tr->alphas[model] = maxTTMinus;	    

      makeGammaCats(tr, model);    
    }
  else
    { 
      tr->alphas[model] = currentTT;
      makeGammaCats(tr, model);
    }    
}

void optimizeAlpha(tree *tr, analdef *adef)
{
 int k;
 
 double currentTT = tr->alphas[0];
 double startLikelihood = tr->likelihood;
 double maxLikelihoodMinus = tr->likelihood;
 double maxLikelihoodPlus = tr->likelihood;
 double maxTTPlus = currentTT;
 double maxTTMinus = currentTT;
 double spacing = 0.5/((double)optimizeAlphaInvocations);
 double tree_likelihood = startLikelihood;
 boolean finish = FALSE;
     
 k = 1;
      
 while(!finish /*tree_likelihood >= maxLikelihoodMinus*/ && (currentTT - spacing * k) > ALPHA_MIN) 
   {      
     tr->alphas[0] = currentTT - spacing * k;     
     makeGammaCats(tr, 0);
     initrav(tr, tr->start);
     initrav(tr, tr->start->back);
     evaluate(tr, tr->start);                   
     
     tree_likelihood = tr->likelihood;
     
     if(tr->likelihood > maxLikelihoodMinus)
       {
	 finish = (fabs(tree_likelihood - maxLikelihoodMinus) < adef->likelihoodEpsilon);

	 maxLikelihoodMinus = tr->likelihood;
	 maxTTMinus = currentTT - spacing * k;
       }    
     else
	finish = TRUE;
        
     k++;
   }
 
 finish = FALSE;
 k = 1;
 tree_likelihood = startLikelihood;
 while(!finish /*tree_likelihood >= maxLikelihoodPlus*/ && ((currentTT + spacing * k) < ALPHA_MAX)) 
   {      
     tr->alphas[0] = currentTT + spacing * k;     
     makeGammaCats(tr, 0);
     initrav(tr, tr->start);
     initrav(tr, tr->start->back);
     evaluate(tr, tr->start);                  
     
     tree_likelihood = tr->likelihood;
     
   
     
     if(tr->likelihood > maxLikelihoodPlus)
       {
	 finish = (fabs(tree_likelihood - maxLikelihoodPlus) < adef->likelihoodEpsilon);
	 
	 maxLikelihoodPlus = tr->likelihood;
	 maxTTPlus = currentTT + spacing * k;
       }	 
     else
       finish = TRUE;

     k++;
   }
 
 if(maxLikelihoodPlus > startLikelihood || maxLikelihoodMinus > startLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{
	  tr->alphas[0] = maxTTPlus;
	}
      else
	{
	  tr->alphas[0] = maxTTMinus;
	}     
      makeGammaCats(tr, 0);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
    }
  else
    { 
      tr->alphas[0] = currentTT;
      makeGammaCats(tr, 0);
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);      
      evaluate(tr, tr->start);
    }    
}

void alterRates(tree *tr, int k, analdef *adef)
{
  int i;
  double granularity = 0.1/((double)optimizeRatesInvocations);
  double bestLikelihood = tr->likelihood, 
    maxLikelihoodMinus = tr->likelihood, maxLikelihoodPlus = tr->likelihood,
    treeLikelihood, originalRate, maxRateMinus, maxRatePlus;
  boolean finish = FALSE;

  i = 1;
  originalRate = tr->initialRates[k];
  maxRateMinus = maxRatePlus = originalRate;
  treeLikelihood = bestLikelihood;

  while(!finish /*(treeLikelihood >= maxLikelihoodMinus)*/ && ((originalRate - granularity * i) > RATE_MIN))
    {     
      tr->initialRates[k] = originalRate - granularity * i;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, 0);     
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
      treeLikelihood = tr->likelihood;          

      if(tr->likelihood > maxLikelihoodMinus)
	{
	  finish = (fabs(treeLikelihood - maxLikelihoodMinus) < adef->likelihoodEpsilon);
	  
	  maxLikelihoodMinus = tr->likelihood;
	  maxRateMinus = originalRate - granularity * i;
	}      
      else
	finish = TRUE;

      i++;
    }

  i = 1;
  treeLikelihood = bestLikelihood;
  tr->initialRates[k] = originalRate;
  finish = FALSE;

  while(!finish /*(treeLikelihood >= maxLikelihoodPlus)*/ && ((originalRate + i * granularity) < RATE_MAX))
    {     
      tr->initialRates[k] = originalRate + i * granularity;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, 0);     
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);      
      treeLikelihood = tr->likelihood;

      

      if(tr->likelihood > maxLikelihoodPlus)
	{
	  finish = (fabs(treeLikelihood - maxLikelihoodPlus) < adef->likelihoodEpsilon);
	  	  
	  maxLikelihoodPlus = tr->likelihood;
	  maxRatePlus = originalRate + granularity * i;
	}
      else
	finish = TRUE;
      i++;
    }
  if(maxLikelihoodPlus > bestLikelihood || maxLikelihoodMinus > bestLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{	  
	  tr->initialRates[k] = maxRatePlus;
	  initReversibleGTR(tr->rdta, tr->cdta, tr, adef, 0);	  
	}
      else
	{
	 tr->initialRates[k] = maxRateMinus;
	 initReversibleGTR(tr->rdta, tr->cdta, tr, adef, 0);	  	
	}     
    
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);         
    }
  else
    {      
      tr->initialRates[k] = originalRate;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, 0);          
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);   
    }

  
}

double alterRatesMULT(tree *tr, int k, analdef *adef, int model)
{
  int i;
  double granularity = 0.1/((double)optimizeRatesInvocations);
  double bestLikelihood, maxLikelihoodMinus, maxLikelihoodPlus,
    treeLikelihood, originalRate, maxRateMinus, maxRatePlus;
  boolean finish = FALSE;
  
  treeLikelihood = partitionLikelihood(tr, model);  

  /*  printf("Partition %d Rate %d %f -> ", model, k, treeLikelihood);*/

  bestLikelihood = treeLikelihood;
  maxLikelihoodMinus = treeLikelihood;
  maxLikelihoodPlus = treeLikelihood;
  
  i = 1;
  originalRate = tr->initialRates[model * 5 + k];
  maxRateMinus = maxRatePlus = originalRate;
   bestLikelihood;

   while(!finish /*(treeLikelihood >= maxLikelihoodMinus)*/ && ((originalRate - granularity * i) > RATE_MAX))
    {
      tr->initialRates[model * 5 + k] = originalRate - granularity * i;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);         

      treeLikelihood = partitionLikelihood(tr, model);
      
      if(treeLikelihood > maxLikelihoodMinus)
	{
	  finish = (fabs(treeLikelihood - maxLikelihoodMinus) < adef->likelihoodEpsilon);
	  
	  maxLikelihoodMinus = treeLikelihood;
	  maxRateMinus = originalRate - granularity * i;
	}      
      else
	finish = TRUE;

      i++;
    }

  i = 1;
  treeLikelihood = bestLikelihood;
  tr->initialRates[model * 5 + k] = originalRate;
  finish = FALSE;

  while(!finish /*(treeLikelihood >= maxLikelihoodPlus)*/ && ((originalRate + i * granularity) < RATE_MAX))
    {
      tr->initialRates[model * 5 + k] = originalRate + i * granularity;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);        		          

      treeLikelihood = partitionLikelihood(tr, model);

     

      if(treeLikelihood > maxLikelihoodPlus)
	{
	  finish = (fabs(treeLikelihood - maxLikelihoodPlus) < adef->likelihoodEpsilon);
	  
	  maxLikelihoodPlus = treeLikelihood;
	  maxRatePlus = originalRate + granularity * i;
	}      
      else
	finish = TRUE;

      i++;
    }

  if(maxLikelihoodPlus > bestLikelihood || maxLikelihoodMinus > bestLikelihood)
    {
      if(maxLikelihoodPlus > maxLikelihoodMinus)
	{	  
	  tr->initialRates[model * 5 + k] = maxRatePlus;
	  initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);	  
	  /*printf("%f\n", maxLikelihoodPlus);*/
	  return maxLikelihoodPlus;
	}
      else
	{
	  tr->initialRates[model * 5 + k] = maxRateMinus;
	  initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);	  	
	  /*printf("%f\n", maxLikelihoodMinus);*/
	  return maxLikelihoodMinus;
	}              
    }
  else
    {      
      tr->initialRates[model * 5 + k] = originalRate;
      initReversibleGTR(tr->rdta, tr->cdta, tr, adef, model);                
      /*printf("%f\n", bestLikelihood);*/
      return bestLikelihood;
    }
}

void optimizeRates(tree *tr, analdef *adef)
{
  int i, model;
  double start = tr->likelihood;

  if(adef->useMultipleModel)
    {
      double *partialLikelihoods = (double *)malloc(tr->NumberOfModels * sizeof(double));      

      for(model = 0; model < tr->NumberOfModels; model++)
	{	  
	  for(i = 0; i < 5; i++)		      
	    partialLikelihoods[model] = alterRatesMULT(tr, i, adef, model);     
	}

      initrav(tr, tr->start);
      initrav(tr, tr->start->back);
      evaluate(tr, tr->start);   

      tr->fracchange = 0;

#ifdef DEBUG_MODEL
      printf("FULL LH %f\n", tr->likelihood);
#endif

      for(model = 0; model < tr->NumberOfModels; model++)
	{
#ifdef DEBUG_MODEL
	  printf("Partial %d %f\n", model, partialLikelihoods[model]);
#endif
	  tr->fracchange += tr->fracchanges[model] * abs(partialLikelihoods[model]);	  
	}
      tr->fracchange /= abs(tr->likelihood);

#ifdef DEBUG_MODEL
      printf("MULT FRACCHANGE %f\n", tr->fracchange);
#endif

      free(partialLikelihoods);
#ifdef DEBUG_MODEL
      printf("Rates Multiple %f -> %f\n", start, tr->likelihood);
#endif      
    }
  else
    {          
#ifdef DEBUG_MODEL
      printf("Rates Single %f -> ", tr->likelihood);
#endif

      for(i = 0; i < 5; i++)		      
	alterRates(tr, i, adef);

#ifdef DEBUG_MODEL
      printf("SINGLE FRACCHANGE %f\n", tr->fracchange);
      printf("%f \n", tr->likelihood);
#endif
    }

  optimizeRatesInvocations++;
}

static int doublecompare2(const void *p1, const void *p2)
{
  double i = *((double *)p1);

  double j = *((double *)p2);
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}

static int catCompare(const void *p1, const void *p2)
{
 rateCategorize *rc1 = (rateCategorize *)p1;
 rateCategorize *rc2 = (rateCategorize *)p2;

  double i = rc1->accumulatedSiteLikelihood;
  double j = rc2->accumulatedSiteLikelihood;
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}




void categorize(tree *tr, rateCategorize *rc)
{
  int i, k, found;
  double temp, diff, min;

  for (i = 0; i < tr->cdta->endsite; i++) 
      {
	temp = tr->cdta->patrat[i];
	found = 0;
	for(k = 0; k < tr->NumberOfCategories; k++)
	  {
	    if(temp == rc[k].rate || (fabs(temp - rc[k].rate) < 0.001))
	      {
		found = 1;
		tr->cdta->rateCategory[i] = k;				
		break;
	      }
	  }
	if(!found)
	  {
	    min = fabs(temp - rc[0].rate);
	    tr->cdta->rateCategory[i] = 0;

	    for(k = 1; k < tr->NumberOfCategories; k++)
	    {
	      diff = fabs(temp - rc[k].rate);
	      if(diff < min)
		{
		  min = diff;
		  tr->cdta->rateCategory[i] = k;
		}
	    }
	  }
      }

  for(k = 0; k < tr->NumberOfCategories; k++)
    tr->cdta->patrat[k] = rc[k].rate; 

}




void optimizeRateCategories(tree *tr, int categorized, int _maxCategories)
{
  int i, k;
  double initialRate, initialLikelihood, v, leftRate, rightRate, leftLH, rightLH, temp, wtemp;   
  double lower_spacing, upper_spacing;
  int maxCategories = _maxCategories;
  double initialLH = tr->likelihood;
  double *oldRat = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *ratStored = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *oldwr = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *oldwr2 = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  double *lhs = (double *)malloc(sizeof(double) * tr->cdta->endsite);
  int *oldCategory = (int *)malloc(sizeof(int) * tr->cdta->endsite);  
  int oldNumber; 
  double epsilon = 0.00001;

  if(optimizeRateCategoryInvocations == 1)
    {
      lower_spacing = 0.5 / ((double)optimizeRateCategoryInvocations);
      upper_spacing = 1.0 / ((double)optimizeRateCategoryInvocations);
    }
  else
    {
      lower_spacing = 0.05 / ((double)optimizeRateCategoryInvocations);
      upper_spacing = 0.1 / ((double)optimizeRateCategoryInvocations);
    }

  if(lower_spacing < 0.001)
    lower_spacing = 0.001;

  if(upper_spacing < 0.001)
    upper_spacing = 0.001;

  optimizeRateCategoryInvocations++;
  
  oldNumber = tr->NumberOfCategories;


  for(i = 0; i < tr->cdta->endsite; i++)
    {    
      oldCategory[i] = tr->cdta->rateCategory[i];
      ratStored[i] = tr->cdta->patratStored[i];    
      oldRat[i] = tr->cdta->patrat[i];
      oldwr[i] =  tr->cdta->wr[i];
      oldwr2[i] =  tr->cdta->wr2[i];
    }

  


 
#pragma omp parallel for private(initialRate, initialLikelihood, leftLH, rightLH, initialLikelihood, leftRate, rightRate, k, v) schedule(dynamic)
  for(i = 0; i < tr->cdta->endsite; i++)
    {    
      tr->cdta->patrat[i] = tr->cdta->patratStored[i];     
      initialRate = tr->cdta->patrat[i];
      
      initialLikelihood = calcPartialLikelihood(tr, i, initialRate);
      
      
      leftLH = rightLH = initialLikelihood;
      leftRate = rightRate = initialRate;
       
      k = 1;
      while((initialRate - k * lower_spacing > 0.0001) && 
	    ((v = calcPartialLikelihood(tr, i, initialRate - k * lower_spacing)) > leftLH) && 
	    (fabs(leftLH - v) > epsilon))
	{	  
	  leftLH = v;
	  leftRate = initialRate - k * lower_spacing;
	  k++;	  
	}      
   
      k = 1;
      while(((v = calcPartialLikelihood(tr, i, initialRate + k * upper_spacing)) > rightLH) &&
	    (fabs(rightLH - v) > epsilon))
	{
	  rightLH = v;
	  rightRate = initialRate + k * upper_spacing;	 
	  k++;
	}           

      if(rightLH > initialLikelihood || leftLH > initialLikelihood)
	{
	  if(rightLH > leftLH)	    
	    {	     
	      tr->cdta->patrat[i] = rightRate;
	      lhs[i] = rightLH;
	    }
	  else
	    {	      
	      tr->cdta->patrat[i] = leftRate;
	      lhs[i] = leftLH;
	    }
	}
      else
	lhs[i] = initialLikelihood;

      tr->cdta->patratStored[i] = tr->cdta->patrat[i];     
      
    }
     

  {     
    rateCategorize *rc = (rateCategorize *)malloc(sizeof(rateCategorize) * tr->cdta->endsite);
    int where;
    int found = 0;
    for (i = 0; i < tr->cdta->endsite; i++)
      {
	rc[i].accumulatedSiteLikelihood = 0;
	rc[i].rate = 0;
      }
      
    where = 1;   
    rc[0].accumulatedSiteLikelihood = lhs[0];/*calcPartialLikelihood(tr, 0, tr->cdta->patrat[0]);*/
    rc[0].rate = tr->cdta->patrat[0];
    tr->cdta->rateCategory[0] = 0;
    
    for (i = 1; i < tr->cdta->endsite; i++) 
      {
	temp = tr->cdta->patrat[i];
	found = 0;
	for(k = 0; k < where; k++)
	  {
	    if(temp == rc[k].rate || (fabs(temp - rc[k].rate) < 0.001))
	      {
		found = 1;						
		rc[k].accumulatedSiteLikelihood += lhs[i];	
		break;
	      }
	  }
	if(!found)
	  {	    
	    rc[where].rate = temp;	    
	    rc[where].accumulatedSiteLikelihood += lhs[i];	    
	    where++;
	  }
	}

    qsort(rc, where, sizeof(rateCategorize), catCompare);
  
    if(where < maxCategories)
      {
	tr->NumberOfCategories = where;
	categorize(tr, rc);
      }
    else
      {
	tr->NumberOfCategories = maxCategories;	
	categorize(tr, rc);
      }
      
    free(rc);
  
    for (i = 0; i < tr->cdta->endsite; i++) 
      {
	/*printf("%d %f\n", i, tr->cdta->patrat[tr->cdta->rateCategory[i]]);*/
	temp = tr->cdta->patrat[tr->cdta->rateCategory[i]];

	tr->cdta->wr[i]  = wtemp = temp * tr->cdta->aliaswgt[i];
	tr->cdta->wr2[i] = temp * wtemp;
      }     

    initrav(tr, tr->start);    
    initrav(tr, tr->start->back);   
    evaluate(tr, tr->start);   

    if(tr->likelihood < initialLH)
      {
	/*printf("No improvement->restoring %f < %f\n", tr->likelihood, initialLH);*/
	tr->NumberOfCategories = oldNumber;
	for (i = 0; i < tr->cdta->endsite; i++)
	    {
	      tr->cdta->patratStored[i] = ratStored[i]; 
	      tr->cdta->rateCategory[i] = oldCategory[i];
	      tr->cdta->patrat[i] = oldRat[i];	    
	      tr->cdta->wr[i]  = oldwr[i];
	      tr->cdta->wr2[i] = oldwr2[i];
	    }       

	initrav(tr, tr->start);
	initrav(tr, tr->start->back);
        evaluate(tr, tr->start);

      }
    }
  
  free(oldCategory);
  free(oldRat);
  free(ratStored);
  free(oldwr);
  free(oldwr2); 
  free(lhs);
}
  

void optimizeAlphas(tree *tr, analdef *adef)
{
  int model;
 
  if(adef->useMultipleModel)
    {

#ifdef DEBUG_MODEL
      printf("Alpha Multiple %f ->", tr->likelihood);
#endif

      for(model = 0; model < tr->NumberOfModels; model++)       
	optimizeAlphaMULT(tr, model, adef);	  
	    
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);    
      evaluate(tr, tr->start);      

#ifdef DEBUG_MODEL
      printf(" %f\n", tr->likelihood);
#endif

    }
  else
    {      

#ifdef DEBUG_MODEL
      printf("Alpha Single %f ->", tr->likelihood);
#endif

      optimizeAlpha(tr, adef);

#ifdef DEBUG_MODEL
      printf(" %f\n", tr->likelihood);
#endif

    }

  optimizeAlphaInvocations++;
}



int optimizeModel (tree *tr, analdef *adef, int finalOptimization)
{
  double startLH = tr->likelihood;
  double initialLH;
  int oldInv;

  switch(adef->model)
    {    
    case M_GTRCAT:
       {	 
	 oldInv = optimizeRatesInvocations;
	 do
	   {	    
	     initialLH = tr->likelihood;
	     optimizeRates(tr, adef);       
	   }
	 while(tr->likelihood > initialLH && optimizeRatesInvocations < oldInv + 10);  	     

	 oldInv = optimizeRateCategoryInvocations;	 
#ifdef DEBUG_MODEL
	 printf("Multiple CAT %f -> ", tr->likelihood);
#endif
	 optimizeRateCategories(tr, finalOptimization, adef->categories);       	
#ifdef DEBUG_MODEL
	 if(tr->NumberOfModels > 1)
	   {
	     int i, j;
	     int **check;
	     check = (int **)malloc(tr->NumberOfModels * sizeof(int *));

	     for(i = 0; i < tr->NumberOfModels; i++)
	       {
		 check[i] = (int *)malloc(tr->NumberOfCategories * sizeof(int));
		 for(j = 0; j < tr->NumberOfCategories; j++)
		   check[i][j] = 0;
	       }

	     for(i = 0; i < tr->cdta->endsite; i++)
	       {
		 check[tr->model[i]][tr->cdta->rateCategory[i]] += 1;
	       }
	     
	     for(i = 0; i < tr->NumberOfModels; i++)
	       {
		 printf("P %d: ", i);
		 for(j = 0; j < tr->NumberOfCategories; j++)
		   printf("%d ", check[i][j]);
		 printf("\n");
	       }
	     
	     for(i = 0; i < tr->NumberOfModels; i++)
	       free(check[i]);

	     free(check);
	   }
	 printf("%f Categories %d\n", tr->likelihood, adef->categories);
#endif


       }                    
      break;    
    case M_GTRGAMMA:
       {
	 oldInv = optimizeRatesInvocations;

	 do
	   {	    
	     initialLH = tr->likelihood;
	     optimizeRates(tr, adef);       	   	    
	   }
	 while((fabs(tr->likelihood - initialLH) > adef->likelihoodEpsilon) && optimizeRatesInvocations < oldInv + 10);       
	 
	 oldInv = optimizeAlphaInvocations;
	
	 do
	   {    	   
	     initialLH = tr->likelihood;
	     optimizeAlphas(tr, adef);		   	  
	   }
	 while((fabs(tr->likelihood - initialLH) > adef->likelihoodEpsilon) && optimizeAlphaInvocations < oldInv + 10);
	 
	 
       }
      break;             
    }

  if(optimizeRatesInvocations > 90)
    optimizeRatesInvocations = 90;  
  if(optimizeRateCategoryInvocations > 90)
    optimizeRateCategoryInvocations = 90;
  if(optimizeAlphaInvocations > 90)
    optimizeAlphaInvocations = 90;

  if(startLH > tr->likelihood) return 0;
  else return 1;
}



